/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rg3DRom.h"
#include "rg3DRom_support.h"
#include <iostream>
#include <string.h>
#include <math.h>

sk3DRomTexture::sk3DRomTexture()
{
	m_flags = 0;
	m_tex_alternate = 0;
}

sk3DRomTextureFace::sk3DRomTextureFace( const int fid, const rgVector a,
	const rgVector b, const rgVector c)
{
	m_fid = fid;
	m_v[0] = a;
	m_v[1] = b;
	m_v[2] = c;
}

int sk3DRomTexture::load()
{
	return load( m_fname.c_str());
}

int sk3DRomTexture::load( const char *p_fname)
{
	// XXX: Texture filename "m_fname" should be preserved if the load
	//	fails.
	int tmp_flags = RGGL_LARGE_TEXTURE;
	if( is_smooth()) tmp_flags |= RGGL_SMOOTH_TEXTURE;
	if( is_mipmapped()) tmp_flags |= RGGL_MIPMAP;
	if( is_alpha()) tmp_flags |= RGGL_COLOR_KEY;
	m_tex.setColorKey( m_colorkey[0], m_colorkey[1], m_colorkey[2]);
	const int err = m_tex.load( p_fname, tmp_flags);
	if( err) return err;
	m_fname = p_fname;
	return 0;
}

void sk3DRomTexture::clear()
{
	m_fname = std::string();	// FIXME: Not all support "::clear()"
	m_flags = 0;
	m_tex.clear();
}

void hlp_read_short_variable_string( std::string &des, std::istream &inf,
	const int len)
{
	des = std::string();	// FIXME: Not all support "::clear()"
	char c;
	for( int j = 0; j < len; j++) {
		inf.read( &c, 1);
		des += c;
	}
}

int sk3DRomTexture::read_file_name( std::istream &inf, const int len)
{
	hlp_read_short_variable_string( m_fname, inf, len);
	if( inf.fail()) return -1;
	return 0;
}

int sk3DRomTexture::read_tex_name( std::istream &inf, const int len)
{
	hlp_read_short_variable_string( m_tname, inf, len);
	if( inf.fail()) return -1;
	return 0;
}

sk3DRomTexturePage::sk3DRomTexturePage()
{
	m_name[0] = '\0';
	m_name[ SK3DROM_TEX_PAGE_NAME_LEN - 1] = '\0';
	m_flags = 0;
	m_tex_unit = 0;
	m_texid = -1;
	m_uscale = 1.0f;
	m_vscale = 1.0f;

	set_page_enabled( true);
}

int sk3DRomTexturePage::read_versionA_page( std::ifstream &inf)
{
	inf.read( m_name, SK3DROM_TEX_PAGE_NAME_LEN);
	m_name[ SK3DROM_TEX_PAGE_NAME_LEN -1] = '\0';
	m_flags = int( read_dword( inf));
	m_tex_unit = int( read_dword( inf));
	m_uscale = read_float( inf);
	m_vscale = read_float( inf);
	m_texid = int( read_dword( inf));
	return read_versionA_faces( inf);
}

int sk3DRomTexturePage::read_versionA_faces( std::ifstream &inf)
{
	m_face.clear();
	const int count = int( read_dword( inf));
	if( count < 0) return -1;
        for( int i = 0; i < count; i++) {
		m_face.push_back( sk3DRomTextureFace());
		(m_face[i]).set_fid( int( read_dword( inf)));
		(m_face[i]).update_vertex( 0, rgVector().read_binary( inf));
		(m_face[i]).update_vertex( 1, rgVector().read_binary( inf));
		(m_face[i]).update_vertex( 2, rgVector().read_binary( inf));
        }
	return 0;
}

void sk3DRomTexturePage::select_vertices( const rgVector n, const rgVector *v,
	const int count)
{
	m_sel_verts.clear();
	for( int i = 0; i < int( m_face.size()); i++) {
		
		// Get the current vertex, project it onto the
		// plane of selection, and check if it is within
		// the selection box.
		if( rgPlanarPointInConvexFace( v, count,
			rgMakePlanar( m_face[i].v( 0), n))) {
			m_sel_verts.push_back(
				sk3DRomTexturePageSelVertex( i, 0));
		}

		if( rgPlanarPointInConvexFace( v, count,
			rgMakePlanar( m_face[i].v( 1), n))) {
			m_sel_verts.push_back(
				sk3DRomTexturePageSelVertex( i, 1));
		}

		if( rgPlanarPointInConvexFace( v, count,
			rgMakePlanar( m_face[i].v( 2), n))) {
			m_sel_verts.push_back(
				sk3DRomTexturePageSelVertex( i, 2));
		}
	}
}

void sk3DRomTexturePage::select_all_vertices()
{
	m_sel_verts.clear();
	for( int i = 0; i < int( m_face.size()); i++) {
		m_sel_verts.push_back( sk3DRomTexturePageSelVertex( i, 0));
		m_sel_verts.push_back( sk3DRomTexturePageSelVertex( i, 1));
		m_sel_verts.push_back( sk3DRomTexturePageSelVertex( i, 2));
	}
}

rgVector sk3DRomTexturePage::get_center_of_selected_vertices() const
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	return (vmin + vmax) * 0.5f;
}

void sk3DRomTexturePage::get_sel_verts_bbox( rgVector &vmin, rgVector &vmax)
	const
{
	vmin = rgVector( 16777816.0f, 16777816.0f, 16777816.0f);
	vmax = rgVector( -16777816.0f, -16777816.0f, -16777816.0f);
	for( int i = 0; i < int( m_sel_verts.size()); i++) {
		const rgVector u( (face( m_sel_verts[i].fid))->v(
			m_sel_verts[i].vid));
		if( u.x < vmin.x) vmin.x = u.x;
		if( u.y < vmin.y) vmin.y = u.y;
		if( u.z < vmin.z) vmin.z = u.z;
		if( u.x > vmax.x) vmax.x = u.x;
		if( u.y > vmax.y) vmax.y = u.y;
		if( u.z > vmax.z) vmax.z = u.z;
	}
}

void sk3DRomTexturePage::scale_selected_to_unit_texture()
{
	scale_selected_to_unit_texture( 0);
}

void sk3DRomTexturePage::scale_selected_to_unit_texture( const float b)
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	scale_selected_to_unit_texture( b, vmin, vmax);
}

void sk3DRomTexturePage::scale_selected_to_unit_texture( const float b,
	const rgVector vmin, const rgVector vmax)
{
	const float length = RG_MAX( vmax.x - vmin.x, vmax.y - vmin.y) +
		(b * 2.0f);
	if( 0.0f == length) return;
	scale_selected( 1.0f / length, vmin);
}

void sk3DRomTexturePage::scale_selected( const float t)
{
	scale_selected( t, t);
}

void sk3DRomTexturePage::scale_selected( const float t,
	const rgVector vmin)
{
	scale_selected( t, t, vmin);
}

void sk3DRomTexturePage::scale_selected( const float u, const float v)
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	scale_selected( u, v, vmin);
}

void sk3DRomTexturePage::scale_selected( const float u, const float v,
	const rgVector vmin)
{
	rgMatrix m;
	m.translate( vmin);
	m.scale( u, v, 1.0f);
	m.translate( -1.0f * vmin);
	transform_selected_vertices( m);
}

void sk3DRomTexturePage::left_justify_selected( const rgVector p)
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	left_justify_selected( p, vmin);
}

void sk3DRomTexturePage::left_justify_selected( const rgVector p,
	const rgVector vmin)
{
	rgMatrix t;
	t.translate( p.x - vmin.x, 0, 0);
	transform_selected_vertices( t);
}

void sk3DRomTexturePage::right_justify_selected( const rgVector p)
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	rgMatrix t;
	t.translate( p.x - vmax.x, 0, 0);
	transform_selected_vertices( t);
}

void sk3DRomTexturePage::top_justify_selected( const rgVector p)
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	rgMatrix t;
	t.translate( 0, p.y - vmax.y, 0);
	transform_selected_vertices( t);
}

void sk3DRomTexturePage::bottom_justify_selected( const rgVector p)
{
	rgVector vmin, vmax;
	get_sel_verts_bbox( vmin, vmax);
	bottom_justify_selected( p, vmin);
}

void sk3DRomTexturePage::bottom_justify_selected( const rgVector p,
	const rgVector vmin)
{
	rgMatrix t;
	t.translate( 0, p.y - vmin.y, 0);
	transform_selected_vertices( t);
}

void sk3DRomTexturePage::select_vertex( const rgVector p, const float radius)
{
	float near_dis = fabs( radius);
	m_sel_vertex = sk3DRomTexturePageSelVertex( -1, -1);
	float d;
	const int fid = select_face( p);
	if( fid >= 0) {
		if( (d = (m_face[fid].v(0) - p).length()) < near_dis) {
			m_sel_vertex.fid = fid;
			m_sel_vertex.vid = 0;
			near_dis = d;
		}

		if( (d = (m_face[fid].v(1) - p).length()) < near_dis) {
			m_sel_vertex.fid = fid;
			m_sel_vertex.vid = 1;
			near_dis = d;
		}

		if( (d = (m_face[fid].v(2) - p).length()) < near_dis) {
			m_sel_vertex.fid = fid;
			m_sel_vertex.vid = 2;
			near_dis = d;
		}
		if( m_sel_vertex.fid >= 0) return;
	}

	for( int i = 0; i < int( m_face.size()); i++) {

		if( (d = (m_face[i].v(0) - p).length()) < near_dis) {
			m_sel_vertex.fid = i;
			m_sel_vertex.vid = 0;
			near_dis = d;
		}

		if( (d = (m_face[i].v(1) - p).length()) < near_dis) {
			m_sel_vertex.fid = i;
			m_sel_vertex.vid = 1;
			near_dis = d;
		}

		if( (d = (m_face[i].v(2) - p).length()) < near_dis) {
			m_sel_vertex.fid = i;
			m_sel_vertex.vid = 2;
			near_dis = d;
		}
	}
}

int sk3DRomTexturePage::select_face( const rgVector p)
{
	tri_sel_cache.start_selection();
	const int istop = int( m_face.size());
	for( int i = 0; i < istop; i++) {
		if( rgPlanarPointInTriangle( m_face[i].v( 0),
			m_face[i].v( 1), m_face[i].v( 2), p)) {
			if( !tri_sel_cache.was_preselected( i)) return i;
		}
	}

	return tri_sel_cache.roll_over_selection();
}

void sk3DRomTexturePage::composite_page( const sk3DRomTexturePage &src)
{
	if( (&src) == this) return;
	float des_xmax = -16777816.0f;
	if( face_count() <= 0) des_xmax = 0.0f;
	else {
		for( int i = 0; i < face_count(); i++) {
			if( ((face( i))->v( 0)).x > des_xmax)
				des_xmax = ((face( i))->v( 0)).x;
	
			if( ((face( i))->v( 1)).x > des_xmax)
				des_xmax = ((face( i))->v( 1)).x;
	
			if( ((face( i))->v( 2)).x > des_xmax)
				des_xmax = ((face( i))->v( 2)).x;
		}
	}

	float src_xmin = 16777816.0f;
	{
		for( int i = 0; i < src.face_count(); i++) {
			if( ((src.face( i))->v( 0)).x < src_xmin)
				src_xmin = ((src.face( i))->v( 0)).x;
	
			if( ((src.face( i))->v( 1)).x < src_xmin)
				src_xmin = ((src.face( i))->v( 1)).x;
	
			if( ((src.face( i))->v( 2)).x < src_xmin)
				src_xmin = ((src.face( i))->v( 2)).x;
		}
	}

	const rgVector xoff( (des_xmax + 0.1f) - src_xmin, 0, 0);
	for( int i = 0; i < src.face_count(); i++) {
		add_face( sk3DRomTextureFace(
				(src.face( i))->fid(),
				(src.face( i))->v( 0) + xoff,
				(src.face( i))->v( 1) + xoff,
				(src.face( i))->v( 2) + xoff));
	}
}

int sk3DRomTexturePage::generate_stencil( rgBMP &bmp, const int dw,
	const int dh) const
{
	if( dw < 2) return -1;
	if( dh < 2) return -2;

	GLint vport[4];
	GLint &vw = vport[2];
	GLint &vh = vport[3];
	glGetIntegerv( GL_VIEWPORT, vport);

	glMatrixMode( GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho( -vw/2.0f, vw/2.0f, -vh/2.0f, vh/2.0f, 1.0f, -1.0f);

	glMatrixMode( GL_MODELVIEW);
	glPushMatrix();

	GLfloat clear_col[4];
	glGetFloatv( GL_COLOR_CLEAR_VALUE, clear_col);
	glClearColor( 1, 1, 1, 1);
	glColor4f( 0, 0, 0, 1);

	GLint pixel_pack_align;
	glGetIntegerv( GL_PACK_ALIGNMENT, &pixel_pack_align);
	glPixelStorei( GL_PACK_ALIGNMENT, 1);

	GLint pixel_unpack_align;
	glGetIntegerv( GL_UNPACK_ALIGNMENT, &pixel_unpack_align);
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1);

	glDisable( GL_DITHER);

	const int ucw = vw - 2;	// Useable copy width
	const int uch = vh - 2; // Useable copy height
	bmp.generate_bmp( dw, dh, 24);
	int y_remain = dh;
	while( y_remain) {
		int x_remain = dw;
		while( x_remain) {
			glLoadIdentity();
			glTranslatef( (-vw/2.0f) + 1, (-vh/2.0f) + 1, 0);
			glTranslatef( -(dw - x_remain), -(dh - y_remain), 0);
			glScalef( float( dw), float( dh), 1.0f);
			glClear( GL_COLOR_BUFFER_BIT);
			glBegin( GL_TRIANGLES);
			render_faces();
			glEnd();
			glFlush();
			glFinish();

			// XXX: When reading the back buffer, remember that
			//	there is a 1 pixel border around the viewport
			//	edges which must not be read since OpenGL
			//	places unwanted clipping edge lines there (that
			//	is it will add edges for polygons that are
			//	clipped by the viewport, thus making the edge
			//	pixels unsuitable for tiling of very large
			//	page stencils).
			rgBMP b;
			if( rggl_read_pixels( b, GL_BACK, 1, 1,
				RG_MIN( x_remain, ucw),
				RG_MIN( y_remain, uch))) return -3;
			if( bmp.copy_sub_image( b, 0, 0, dw - x_remain,
				dh - y_remain, b.getWidth(), b.getHeight()))
				return -4;
			if( x_remain > ucw) x_remain -= ucw;
			else x_remain = 0;
		}

		if( y_remain > uch) y_remain -= uch;
		else y_remain = 0;
	}

	glPopMatrix();
	glMatrixMode( GL_PROJECTION);
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW);

	glPixelStorei( GL_PACK_ALIGNMENT, pixel_pack_align);
	glPixelStorei( GL_UNPACK_ALIGNMENT, pixel_unpack_align);
	glClearColor( clear_col[0], clear_col[1], clear_col[2], clear_col[3]);

	return 0;
}

int sk3DRomModel::add_texture()
{
	const int tid = texture_count();
	texture.push_back( sk3DRomTexture());
	return tid;
}

int sk3DRomModel::add_page()
{
	const int pid = page_count();
	m_page.push_back( sk3DRomTexturePage());
	return pid;
}

void sk3DRomModel::notify_pages_of_face_delete( const int fid)
{
	for( int i = 0; i < page_count(); i++) {
		for( int j = 0; j < (m_page[i]).face_count(); j++) {
			if( fid == ((m_page[i]).face( j))->fid()) {
				m_page[i].delete_face( j);
				return;
			}
		}
	}
}

void sk3DRomModel::notify_pages_of_face_swap( const int old_fid,
	const int new_fid)
{
	for( int i = 0; i < page_count(); i++) {
		for( int j = 0; j < (m_page[i]).face_count(); j++) {
			if( old_fid == ((m_page[i]).face( j))->fid()) {
				m_page[i].change_face_fid( j, new_fid);
				return;
			}
		}
	}
}

void sk3DRomModel::notify_pages_of_texture_delete( const int tid)
{
	for( int i = 0; i < page_count(); i++)
		if( tid == m_page[i].texid()) m_page[i].bind_texture( -1);
}

void sk3DRomModel::notify_pages_of_texture_swap( const int old_tid,
	const int new_tid)
{
	for( int i = 0; i < page_count(); i++) {
		if( old_tid == m_page[i].texid())
			m_page[i].bind_texture( new_tid);
	}
}

void sk3DRomModel::delete_texture( const int tid)
{
	if( (tid < 0) || (tid >= texture_count())) return;

	const int last = texture_count() - 1;
	notify_pages_of_texture_delete( tid);
	notify_pages_of_texture_swap( last, tid);
	texture[tid] = texture[last];
	texture.erase( texture.begin() + last);
}

void sk3DRomModel::delete_page( const int pid)
{
	if( (pid < 0) || (pid >= page_count())) return;
	m_page.erase( m_page.begin() + pid);
}

void sk3DRomModel::render_page( const int pid) const
{
	if( (pid < 0) || (pid >= page_count())) return;

	const sk3DRomTexturePage &p = m_page[ pid];
	const int tid = p.texid();
	if( (tid >= 0) && (tid < texture_count())) {
		const sk3DRomTexture &t = texture[ tid];
		t.bind();
		if( p.world_texture()) {
			// FIXME: Should use actual viewable dimensions rather
			//	than a simple large quad for "world texture".
			//	Also, using a large number causes precision
			//	problems with float/double numbers.
			glBegin( GL_QUADS);
			//const double k = 65536.0f;
			const double k = 16384.0f;
			const double ku = p.uscale();
			const double kv = p.vscale();
			glTexCoord2d( -k * ku, k * kv);
			glVertex2d( -k, k);

			glTexCoord2d( k * ku, k * kv);
			glVertex2d( k, k);

			glTexCoord2d( k * ku, -k * kv);
			glVertex2d( k, -k);

			glTexCoord2d( -k * ku, -k * kv);
			glVertex2d( -k, -k);
			glEnd();

		} else {
			glBegin( GL_QUADS);
			glTexCoord2f( 0, p.vscale());
			glVertex2f( 0, 1);
			//glVertex2f( 0, t.height());

			glTexCoord2f( p.uscale(), p.vscale());
			glVertex2f( 1, 1);
			//glVertex2f( t.width(), t.height());

			glTexCoord2f( p.uscale(), 0);
			glVertex2f( 1, 0);
			//glVertex2f( t.width(), 0);

			glTexCoord2f( 0, 0);
			glVertex2f( 0, 0);
			glEnd();
		}

	}


	glBindTexture( GL_TEXTURE_2D, 0);
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
	glEnable( GL_CULL_FACE);
	glCullFace( GL_FRONT);

	glBegin( GL_TRIANGLES);
	glColor3f( 1, 0, 0);
	m_page[ pid].render_faces();
	glEnd();

	glEnable( GL_LINE_STIPPLE);
	glLineStipple( 1, 0xF0F0);
	glBegin( GL_TRIANGLES);
	glColor3f( 0, 1, 0);
	m_page[ pid].render_faces();
	glEnd();

	glDisable( GL_LINE_STIPPLE);
	glCullFace( GL_BACK);
	glBegin( GL_TRIANGLES);
	glColor3f( 0, 0, 1);
	m_page[ pid].render_faces();
	glEnd();
	glDisable( GL_CULL_FACE);


	// XXX: Box border
	glColor3f( 0, 0, 0);
	glBegin( GL_QUADS);
	glVertex2f( 0, 1);
	glVertex2f( 1, 1);
	glVertex2f( 1, 0);
	glVertex2f( 0, 0);
	glEnd();
	glLineStipple( 1, 0xF0F0);
	glEnable( GL_LINE_STIPPLE);
	glBegin( GL_QUADS);
	glColor3f( 1, 1, 0);
	glVertex2f( 0, 1);
	glVertex2f( 1, 1);
	glVertex2f( 1, 0);
	glVertex2f( 0, 0);
	glEnd();
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
	glDisable( GL_LINE_STIPPLE);
}

void sk3DRomModel::add_tv_face_to_page( const int pid, const int fid,
	const rgVector a, const rgVector b, const rgVector c)
{
	if( (pid < 0) || (pid >= page_count())) return;
	if( (fid < 0) || (fid >= face_count())) return;
	m_page[ pid].add_face( fid, a, b, c);
}

void sk3DRomModel::compile_texture_data()
{
	for( int i = 0; i < int( page_count()); i++)
		compile_page_texture_data( i);
}

void sk3DRomModel::compile_page_texture_data( const int pid)
{
	if( (pid < 0) || (pid >= page_count())) return;
	sk3DRomTexturePage &p = m_page[ pid];
	if( !(p.page_enabled())) return;
	for( int i = 0; i < p.face_count(); i++) {
		const sk3DRomTextureFace *pf = p.face( i);	// Page Face
		sk3DRomFace &mf = f[ pf->fid()];	// Model Face
		mf.tid = p.texid();
		mf.map[0] = (pf->v(0)).x * p.uscale();
		mf.map[1] = (pf->v(0)).y * p.vscale();

		mf.map[2] = (pf->v(1)).x * p.uscale();
		mf.map[3] = (pf->v(1)).y * p.vscale();

		mf.map[4] = (pf->v(2)).x * p.uscale();
		mf.map[5] = (pf->v(2)).y * p.vscale();
	}
}

void sk3DRomFace::gen_planar_map( const sk3DRomModel &mdl,
	const rgCoordSystem &cs)
{
	const rgVector a( rgMakePlanar( mdl.get_tv_vertex( vid[0]), cs.eye)
		- cs.p);
	const rgVector b( rgMakePlanar( mdl.get_tv_vertex( vid[1]), cs.eye)
		- cs.p);
	const rgVector c( rgMakePlanar( mdl.get_tv_vertex( vid[2]), cs.eye)
		- cs.p);
	map[0] = a.dot( cs.side);
	map[1] = a.dot( cs.up);
	map[2] = b.dot( cs.side);
	map[3] = b.dot( cs.up);
	map[4] = c.dot( cs.side);
	map[5] = c.dot( cs.up);
}

void sk3DRomModel::gen_planar_map( bool p_visible, const rgMatrix &m,
	const int pid)
{
	rgCoordSystem cs;
	cs.side = m * cs.side;
	cs.up = m * cs.up;
	cs.eye = m * cs.eye;
	gen_planar_map( p_visible, cs, pid);
}

void sk3DRomModel::gen_planar_map( bool p_visible, const rgCoordSystem &cs,
	const int pid)
{
	if( (pid < 0) || (pid >= page_count())) return;
	for( int i = 0; i < face_count(); i++) {
		if( !p_visible || !(face_hidden( i))) {
			(f[i]).gen_planar_map( *this, cs);
			(m_page[ pid]).add_face( i,
				rgVector( f[i].map[0], f[i].map[1], 0),
				rgVector( f[i].map[2], f[i].map[3], 0),
				rgVector( f[i].map[4], f[i].map[5], 0));
		}
	}
}

void sk3DRomModel::set_group_of_visible_faces( int gid)
{
	if( (gid < 0) || (gid >= group_count())) gid = -1;
	for( int i = 0; i < face_count(); i++)
		if( !face_hidden( i)) f[i].gid = gid;
}

void sk3DRomModel::set_color_of_face_group( const int gid, float r, float g,
	float b)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++) {
		if( gid == f[i].gid) {
			f[i].li[0] = (unsigned char)(r * 255.0f);
			f[i].li[1] = (unsigned char)(g * 255.0f);
			f[i].li[2] = (unsigned char)(b * 255.0f);

			f[i].li[4] = (unsigned char)(r * 255.0f);
			f[i].li[5] = (unsigned char)(g * 255.0f);
			f[i].li[6] = (unsigned char)(b * 255.0f);

			f[i].li[8] = (unsigned char)(r * 255.0f);
			f[i].li[9] = (unsigned char)(g * 255.0f);
			f[i].li[10] = (unsigned char)(b * 255.0f);
		}
	}
}

void sk3DRomModel::set_alpha_of_face_group( const int gid, float a)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++) {
		if( gid == f[i].gid) {
			f[i].li[3] = (unsigned char)(a * 255.0f);
			f[i].li[7] = (unsigned char)(a * 255.0f);
			f[i].li[11] = (unsigned char)(a * 255.0f);
		}
	}
}

void sk3DRomModel::set_use_color_of_face_group( const int gid)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++)
		if( gid == f[i].gid) f[i].set_use_color();
}

void sk3DRomModel::unset_use_color_of_face_group( const int gid)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++)
		if( gid == f[i].gid) f[i].unset_use_color();
}

void sk3DRomModel::force_flat_face_group( const int gid)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++)
		if( gid == f[i].gid) f[i].force_flat_shade();
}

void sk3DRomModel::unforce_flat_face_group( const int gid)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++)
		if( gid == f[i].gid) f[i].unforce_flat_shade();
}

void sk3DRomModel::set_bsp_state_for_face_group( const int gid, bool p_state)
{
	if( (gid < 0) || (gid >= group_count())) return;
	for( int i = 0; i < face_count(); i++) {
		if( gid == f[i].gid) {
			if( p_state) f[i].unset_bsp();
			else f[i].set_bsp();
		}
	}
}

void sk3DRomModel::scale_selected_page_faces_to_unit_texture( const int pid,
	const float p_border)
{
	if( (pid < 0) || (pid >= page_count())) return;
	if( m_page[pid].texid() < 0) return;
	if( p_border < 0) return;
	rgVector vmin, vmax;
	(m_page[pid]).get_sel_verts_bbox( vmin, vmax);
	const float b = RG_MAX(
		p_border / float( texture[ m_page[pid].texid()].width()),
		p_border / float( texture[ m_page[pid].texid()].height()));
	(m_page[pid]).scale_selected_to_unit_texture( b, vmin, vmax);
	(m_page[pid]).left_justify_selected( rgVector( b, 0, 0), vmin);
	(m_page[pid]).bottom_justify_selected( rgVector( 0, b, 0), vmin);
}
