/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGPHYSICS_H
#define RGPHYSICS_H

#include <rgStandard.h>
#include <vector>

class rgCollisionPrisim;

class rgCollisionPlane
{
public:
	rgCollisionPlane( const rgVector p, const rgVector n) {
		m_point = p; m_normal = n;}
	rgVector m_point;
	rgVector m_normal;
};

class rgCollisionQuad
{
public:
	rgCollisionQuad(){}
	rgCollisionQuad( const rgVector a, const rgVector b, const rgVector c,
		const rgVector d) {
		m_v[0] = a; m_v[1] = b; m_v[2] = c; m_v[3] = d;
		m_normal = (rgCalcTriangleNormal( a, b, c)).normalize();
	}

	void render() const;
	void render_normal( const float k) const;

	rgVector normal() const { return m_normal;}
	rgVector center() const { return ((m_v[0] + m_v[2]) * 0.5f);}

	bool overlaps( const rgCollisionQuad &cq) const;


	bool contains_quad_points( const rgCollisionQuad &cq) const {
		if( rgPlanarPointInQuad( m_v, cq.m_v[0])) return true;
		if( rgPlanarPointInQuad( m_v, cq.m_v[1])) return true;
		if( rgPlanarPointInQuad( m_v, cq.m_v[2])) return true;
		if( rgPlanarPointInQuad( m_v, cq.m_v[3])) return true;
		return false;
	}

	bool splits_quad_edges( const rgCollisionQuad &cq) const {
		#define ISECT( a, b, c, d) rgLineSegLineSegIntersect(a,b,c,d)
		if( ISECT( m_v[0], m_v[1], cq.m_v[0], cq.m_v[1])) return true;
		if( ISECT( m_v[0], m_v[1], cq.m_v[1], cq.m_v[2])) return true;
		if( ISECT( m_v[0], m_v[1], cq.m_v[2], cq.m_v[3])) return true;
		if( ISECT( m_v[0], m_v[1], cq.m_v[3], cq.m_v[0])) return true;

		if( ISECT( m_v[1], m_v[2], cq.m_v[0], cq.m_v[1])) return true;
		if( ISECT( m_v[1], m_v[2], cq.m_v[1], cq.m_v[2])) return true;
		if( ISECT( m_v[1], m_v[2], cq.m_v[2], cq.m_v[3])) return true;
		if( ISECT( m_v[1], m_v[2], cq.m_v[3], cq.m_v[0])) return true;

		if( ISECT( m_v[2], m_v[3], cq.m_v[0], cq.m_v[1])) return true;
		if( ISECT( m_v[2], m_v[3], cq.m_v[1], cq.m_v[2])) return true;
		if( ISECT( m_v[2], m_v[3], cq.m_v[2], cq.m_v[3])) return true;
		if( ISECT( m_v[2], m_v[3], cq.m_v[3], cq.m_v[0])) return true;

		if( ISECT( m_v[3], m_v[0], cq.m_v[0], cq.m_v[1])) return true;
		if( ISECT( m_v[3], m_v[0], cq.m_v[1], cq.m_v[2])) return true;
		if( ISECT( m_v[3], m_v[0], cq.m_v[2], cq.m_v[3])) return true;
		if( ISECT( m_v[3], m_v[0], cq.m_v[3], cq.m_v[0])) return true;
		#undef ISECT
		return false;
	}

	void flatten( const rgVector n) {
		m_v[0] = rgMakePlanar( m_v[0], n);
		m_v[1] = rgMakePlanar( m_v[1], n);
		m_v[2] = rgMakePlanar( m_v[2], n);
		m_v[3] = rgMakePlanar( m_v[3], n);
	}

	bool splits_segment( const rgVector a, const rgVector b) const {
		const float t = m_normal.dot( a - m_v[0]);
		const float k = m_normal.dot( b - m_v[1]);
		if( t * k >= 0) return false;
		return true;
	}

	rgVector m_v[4];
	rgVector m_normal;

};

// FIXME: Currently rgCollisionShadow only handles shadows created by
//	rectangular prisims. Later, add support for other volumes.
class rgCollisionShadow
{
public:
	//rgCollisionShadow() { m_qcount = 0;}
	rgCollisionShadow() { m_vcount = 0;}
	rgCollisionShadow( const rgCollisionPrisim &cp, const rgVector n);

	void draw() const;

	bool overlaps( const rgCollisionShadow &cs) const;
	bool span_overlaps( const rgCollisionShadow &ics,
		const rgCollisionShadow &fcs) const;
	//bool rotate_span_overlaps(
	//	const rgCollisionShadow &ics,
	//	const rgCollisionShadow &fcs,
	//	const rgVector p_center) const;
	//bool linear_span_overlaps(
	//	const rgCollisionShadow &ics,
	//	const rgCollisionShadow &fcs,
	//	const rgVector icen,
	//	const rgVector fcen) const;

	rgVector center() const { return m_center;}

	//int qcount() const { return m_qcount;}

protected:

	bool contains_shadow_points( const rgCollisionShadow &cs) const {
		#define PPICF rgPlanarPointInConvexFace
		switch( cs.m_vcount) {
		case 6:
			if( PPICF( m_v, m_vcount, cs.m_v[5])) return true;
			if( PPICF( m_v, m_vcount, cs.m_v[4])) return true;
		case 4:
			if( PPICF( m_v, m_vcount, cs.m_v[3])) return true;
			if( PPICF( m_v, m_vcount, cs.m_v[2])) return true;
			if( PPICF( m_v, m_vcount, cs.m_v[1])) return true;
			if( PPICF( m_v, m_vcount, cs.m_v[0])) return true;
		}
		#undef PPICF

#ifdef asfsdkffj32j98fh2nbc894b8f87ytgbd8g327fb27cb234r7b
		switch( m_vcount) {
		case 6:
			#define PPIH rgPlanarPointInHexagon
			switch( cs.m_vcount) {
			case 6:
				if( PPIH( m_v, cs.m_v[5])) return true;
				if( PPIH( m_v, cs.m_v[4])) return true;
			case 4:
				if( PPIH( m_v, cs.m_v[3])) return true;
				if( PPIH( m_v, cs.m_v[2])) return true;
				if( PPIH( m_v, cs.m_v[1])) return true;
				if( PPIH( m_v, cs.m_v[0])) return true;
			}
			#undef PPIH
			break;
		
		case 4:
			#define PPIQ rgPlanarPointInQuad
			switch( cs.m_vcount) {
			case 6:
				if( PPIQ( m_v, cs.m_v[5])) return true;
				if( PPIQ( m_v, cs.m_v[4])) return true;
			case 4:
				if( PPIQ( m_v, cs.m_v[3])) return true;
				if( PPIQ( m_v, cs.m_v[2])) return true;
				if( PPIQ( m_v, cs.m_v[1])) return true;
				if( PPIQ( m_v, cs.m_v[0])) return true;
			}
			#undef PPIQ
			break;

		}
#endif
		return false;
	}

	bool splits_shadow_edges( const rgCollisionShadow &cs) const;

	bool add_shadow_quad( const rgVector qn, const rgVector n,
		const char p_mask, char &p_flag) {
		//if( m_qcount >= 3) return false;
		if( n.dot( qn) > 0.00001f) {
			//(m_quad[ m_qcount++] = q).flatten( n);
			p_flag |= p_mask;
			return true;
		}
		return false;
	}

	
	void set_verts( const rgCollisionPrisim &cp,
		short a, short b, short c, short d, const rgVector n);
	void set_verts( const rgCollisionPrisim &cp,
		short a, short b, short c, short d, short e, short f,
		const rgVector n);
	//
	//int m_qcount;
	char m_vcount;
	rgVector m_v[6];
	rgVector m_center;
};

class rgCollisionPrisim
{
public:
	rgCollisionPrisim() { create( 1, 1, 1);}

	// Prisim definition:
	// XXX: Without the gaurd bars on the right, the drawing would produce
	//	a "multi-line comment" warning/error.
	//                  |
	// 0-----1          |
	// |\    |\         |
	// | 3-----2        |
	// | |   | |        |
	// 4-|---7 |        |
	//  \|    \|        |
	//   5-----6        |
	//
	// Point 4 is at the origin point (0, 0, 0);
	rgCollisionPrisim &create( float x_scale, float y_scale,
		float z_scale);

	rgCollisionPrisim &create( const std::vector <rgVector> &v,
		const rgMatrix &m);

	rgVector plane_point_A() const { return m_v[0];}
	rgVector plane_point_B() const { return m_v[6];}
	rgVector plane_point_C() const { return m_v[6];}
	rgVector plane_point_D() const { return m_v[0];}
	rgVector plane_point_E() const { return m_v[0];}
	rgVector plane_point_F() const { return m_v[6];}

	rgVector normal_A() const { return (m_v[0] - m_v[4]) * inv_ylen;}
	rgVector normal_B() const { return (m_v[1] - m_v[0]) * inv_xlen;}
	rgVector normal_C() const { return (m_v[4] - m_v[0]) * inv_ylen;}
	rgVector normal_D() const { return (m_v[0] - m_v[1]) * inv_xlen;}
	rgVector normal_E() const { return (m_v[4] - m_v[5]) * inv_zlen;}
	rgVector normal_F() const { return (m_v[5] - m_v[4]) * inv_zlen;}

	bool plane_splits_segment( const rgVector pp, const rgVector pn,
		const rgVector a, const rgVector b) const {
		if( (pn.dot( a - pp) * pn.dot( b - pp)) >= 0) return false;
		return true;
	}

	// 4----7
	// | E  |
	// |    |
	// 0----1----7----4----0
	// | A  | B  | C  | D  |
	// |    |    |    |    |
	// 3----2----6----5----3
	// | F  |
	// |    |
	// 5----6
	rgCollisionQuad quad_A() const {
	        return rgCollisionQuad( m_v[0], m_v[1], m_v[2], m_v[3]);}

	rgCollisionQuad quad_B() const {
	        return rgCollisionQuad( m_v[1], m_v[7], m_v[6], m_v[2]);}

	rgCollisionQuad quad_C() const {
	        return rgCollisionQuad( m_v[7], m_v[4], m_v[5], m_v[6]);}

	rgCollisionQuad quad_D() const {
	        return rgCollisionQuad( m_v[4], m_v[0], m_v[3], m_v[5]);}

	rgCollisionQuad quad_E() const {
	        return rgCollisionQuad( m_v[4], m_v[7], m_v[1], m_v[0]);}

	rgCollisionQuad quad_F() const {
	        return rgCollisionQuad( m_v[3], m_v[2], m_v[6], m_v[5]);}

	//rgCollisionPlane plane_A() const {
	//	return rgCollisionPlane( m_v[1], m_n[0]);}
	//rgCollisionPlane plane_B() const {
	//	return rgCollisionPlane( m_v[1], m_n[1]);}
	//rgCollisionPlane plane_C() const {
	//	return rgCollisionPlane( m_v[7], m_n[2]);}
	//rgCollisionPlane plane_D() const {
	//	return rgCollisionPlane( m_v[3], m_n[3]);}
	//rgCollisionPlane plane_E() const {
	//	return rgCollisionPlane( m_v[7], m_n[4]);}
	//rgCollisionPlane plane_F() const {
	//	return rgCollisionPlane( m_v[3], m_n[5]);}



	rgCollisionPrisim &transform( const rgMatrix &m) {
		m_v[0] = m.pointMul( m_v[0]);
		m_v[1] = m.pointMul( m_v[1]);
		m_v[2] = m.pointMul( m_v[2]);
		m_v[3] = m.pointMul( m_v[3]);
		m_v[4] = m.pointMul( m_v[4]);
		m_v[5] = m.pointMul( m_v[5]);
		m_v[6] = m.pointMul( m_v[6]);
		m_v[7] = m.pointMul( m_v[7]);
//		m_n[0] = m * m_n[0];
//		m_n[1] = m * m_n[1];
//		m_n[2] = m * m_n[2];
//		m_n[3] = m * m_n[3];
//		m_n[4] = m * m_n[4];
//		m_n[5] = m * m_n[5];
		return *this;
	}

	rgCollisionPrisim &translate( float x, float y, float z) {
		return translate( rgVector( x, y, z));
	}
	rgCollisionPrisim &translate( const rgVector u) {
		rgMatrix m;
		m.translate( u);
		return transform( m);
	}

	rgCollisionPrisim &rotate( const float ang, const rgVector v) {
		rgMatrix m;
		m.translate( center());
		m.rotate( ang, v);
		m.translate( -1.0f * center());
		return transform( m);
	}

	rgVector center() const { return (m_v[0] + m_v[6]) * 0.5f;}

	rgCollisionShadow collision_shadow( const rgVector n) const;

	void render() const;
	void render_normals( const float k) const {
		(quad_A()).render_normal( k);
		(quad_B()).render_normal( k);
		(quad_C()).render_normal( k);
		(quad_D()).render_normal( k);
		(quad_E()).render_normal( k);
		(quad_F()).render_normal( k);
	}

	bool collision( const rgCollisionPrisim &tar) const;
	static bool collision( const rgCollisionPrisim &dyn,
		const rgCollisionPrisim &sta);
	static bool collision( const rgCollisionPrisim &dyn,
		const rgCollisionShadow &staty,
		const rgCollisionShadow &statz);

	bool collision( const rgCollisionPlane &tar) const;

	bool collision( const rgVector p) const;
	bool collision( const rgVector p, const float radius) const;

	bool collision(
		const rgVector a,
		const rgVector b,
		const float radius,
		rgVector &surface_normal,
		rgVector &surface_point,
		rgVector &intersect_point,
		rgVector &intrusion_vector) const;

	rgVector collision_surface_normal( const rgVector src,
		const rgVector des) const {
		#define PSS plane_splits_segment
		rgVector n( normal_A());
		if( PSS( plane_point_A(), n, src, des)) return n;
		if( PSS( plane_point_B(), n = normal_B(), src, des)) return n;
		if( PSS( plane_point_C(), n = normal_C(), src, des)) return n;
		if( PSS( plane_point_D(), n = normal_D(), src, des)) return n;
		if( PSS( plane_point_E(), n = normal_E(), src, des)) return n;
		// FIXME: Assuming that the segment MUST collide some surface
		//	of the prisim.
		return normal_F();
		#undef pss
	}

	void bounding_box( rgVector &vmin, rgVector &vmax) const;

//	void refresh_normals();

//	bool intersect( const rgCollisionPrisim &cp) const;

	rgVector m_v[8];
//	rgVector m_n[6];

	float inv_xlen;
	float inv_ylen;
	float inv_zlen;

	friend std::ostream &operator <<( std::ostream &o,
		const rgCollisionPrisim &p);

protected:
	static bool collision( const rgCollisionPrisim &des,
		const rgCollisionPrisim &tar, const rgVector n);
};


class rgPhyBoundSphere
{
public:
	rgVector center;
	float radius;

	rgPhyBoundSphere(){ radius = 1.0f;}
	rgPhyBoundSphere( const rgVector p, const float r) {
		center = p; radius = r;}

	//std::vector <rgPhyBoundSphere> m_sphere;

	rgPhyBoundSphere &gen_containing(
		const std::vector <rgCollisionPrisim> &p);
	rgPhyBoundSphere &gen_containing(
		const std::vector <rgPhyBoundSphere> &s);

};

class rgPhyObject
{
public:
	rgCoordSystem csys;

	// Total mass of object.
	virtual float mass() { return 0;}

	virtual rgVector center_of_mass() { return csys.p;}

};


#endif

