/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <math.h>
#include "rgRay.h"
#include "rgPort.h"

using namespace std;

rgRayScene::rgRayScene()
{
	outFileName = 0;
	viewWidth = 0;
	viewHeight = 0;
	traceDepth = 0;

	lightCount = 0;
	sphereCount = 0;
	triCount = 0;
	faceCount = 0;

	light = 0;
	sphere = 0;
	tri = 0;
	face = 0;
}

rgRayEntity::rgRayEntity()
{
	uid = rgLONG( this);
}

void rgRayTriangle::genNormal()
{
	n = ((b - a).cross( a - c)).normalize();

}

rgLONG rgRayScene::getSphereID( rgRaySphere *p)
{

	const rgLONG id = rgLONG(p - sphere);


	if( id < 0) return -1;
	if( id >= sphereCount) return -1;

//	cerr << "-> " << p << ", " << sphere << ", " << rgLONG(p - sphere) << "\n";
//	cerr << "id = " << id << "\n";
//	exit(0);
	
	return id;
}

rgRayFrameBuffer::rgRayFrameBuffer()
{
	width = 0;
	height = 0;
	colour = 0;
}

int rgRayFrameBuffer::create( const int p_width, const int p_height)
{
	if( p_width <= 0) return -1;
	if( p_height <= 0) return -1;

	width = p_width;
	height = p_height;

	/* Allocate ram for an RGB colour buffer. */
	colour = new rgBYTE[ width * height * 3];

	return 0;
}

rgVector rgRayCamera::getRayForPixel( const int p_x, const int p_y)
{
	rgVector r;

	if( p_x < 0) return r;
	if( p_y < 0) return r;
	if( p_x >= viewXResolution) return r;
	if( p_y >= viewYResolution) return r;

	r.x = float(p_x) - (float(viewXResolution) * 0.5f);
	r.y = -(float(p_y) - (float(viewYResolution) * 0.5f));
	r.z = -viewPlaneDistance;

	r.normalize();
	return r;
}

#define COLOUR_BUF( mx, my, c) (frame.colour[ ((my) * frame.width + (mx))*3 + \
	c])
void rgRayTraceScene(
	rgFloat viewAngle,
	rgFloat aspect,
        rgFloat pnear,
        rgFloat pfar,
	rgRayFrameBuffer &frame,
	rgRayScene &scene)
{

	rgMatrix proj;
        proj.perspective( viewAngle, aspect, pnear, pfar);


	rgVector cop = proj.pointMul( rgVector( 1.0f, 1.0f, 0.0f));
	cerr << "proj.d[15] = " << proj.d[15] << "\n";
	
	cop.x *= (1.0f/cop.z);
	cop.y *= (1.0f/cop.z);
	cerr << "COP is " << cop << "\n";

	rgVector r;
/*
	r = rgVector( 1.0f, 1.0f, 0.0001f);
	r = proj.pointMul( r);

	r.x *= (1.0f/r.z);
	r.y *= (1.0f/r.z);
	cerr << "Ray is " << r << "\n";
	*/


	rgRayCamera cam;
	cam.fieldOfView = RG_DEGTORAD( viewAngle);
	cam.viewXResolution = frame.width;
	cam.viewYResolution = frame.height;

	cam.viewPlaneDistance =
		(rgFloat(cam.viewXResolution >> 1)/
		tan( cam.fieldOfView * 0.5f));

	cout << "View point: " << cam.position << "\n";
	cout << "View rotaion (X,Y,Z): " << cam.rotation << "\n";
	cout << "Feild of view: " << cam.fieldOfView << "\n";
	cout << "View plane X res: " << cam.viewXResolution << "\n";
	cout << "View plane Y res: " << cam.viewYResolution << "\n";
	cout << "View plane distance: " << cam.viewPlaneDistance << "\n";
	cout << "\n";


	r = cam.getRayForPixel( frame.width/2, frame.height/2);
	cerr << "Ray for pixel (" << (frame.width/2) <<
		", " << (frame.height/2) << ") is " << r << "\n";

	int x;
	int y;
	bool hit;
	rgVector col;

	for( y = 0; y < frame.height; y++) {
		for( x = 0; x < frame.width; x++) {

			r = cam.getRayForPixel( x, y);
			hit = rgRay_trace( cam.position, r, scene, col, -1,
				RGRAY_MAX_TRACE_RECURSE);

			if( !hit) {
				/* Nothing traced */
				col = scene.backColour;
			}

			/* Colour fragment should be valid. */
			col.x = RG_MAX( 0.0f, RG_MIN( 1.0f, col.x));
			col.y = RG_MAX( 0.0f, RG_MIN( 1.0f, col.y));
			col.z = RG_MAX( 0.0f, RG_MIN( 1.0f, col.z));
			col = col * 255.0f;

			/* Write the colour fragment to the cbuffer. */
			COLOUR_BUF( x, y, 0) = rgBYTE( col.x);	/* Red */
			//COLOUR_BUF( x, y, 0) = rgBYTE(-cp.z); /* Red */
			//COLOUR_BUF( x, y, 1) = rgBYTE(-cp.z); /* Green */
			//COLOUR_BUF( x, y, 2) = rgBYTE(-cp.z); /* Blue */
			COLOUR_BUF( x, y, 1) = rgBYTE( col.y);	/* Green */
			COLOUR_BUF( x, y, 2) = rgBYTE( col.z);	/* Blue */
		
		}
	}
}

bool rgRay_trace( rgVector rp, rgVector rd, rgRayScene &scene,
	rgVector &colour, rgLONG p_uid_to_skip, int depth)
{
	if( depth <= 0) return false;

	bool hit = false;
	int sid = -1;
	int tid = -1;
	rgFloat dis = 32767.0f;
	rgVector sip;
	rgVector tip;
	rgVector fragPos;
	rgVector fragNorm;
	rgRayColourMaterial fragMat;
	rgRayObject *obj = 0;
	rgLONG uid = 0;
	int fragObj = RGRAY_NO_OBJECT;

	/***************************************************************
	 * Find object visible with the curent ray.
	 ***************************************************************/
	rgRayFindClosestSphere( scene, rp, rd, sid, sip, p_uid_to_skip);
	rgRayFindClosestTriangle( scene, rp, rd, tid, tip, p_uid_to_skip);

	if( sid >= 0) {
		if( (rp - sip).length() < dis) {
			dis = (rp - sip).length();
			fragObj = RGRAY_SPHERE;
		}
	}

	if( tid >= 0) {
		if( (rp - tip).length() < dis) {
			dis = (rp - tip).length();
			fragObj = RGRAY_TRIANGLE;
		}
	}
	
	switch( fragObj) {
	case RGRAY_SPHERE:
		/* Fragment is from a sphere */
		obj = &scene.sphere[ sid];
		uid = scene.sphere[sid].uid;
		fragPos = sip;
		fragNorm = sip - scene.sphere[sid].center;
		fragNorm.normalize();
		fragMat.amb = scene.sphere[sid].amb;
		fragMat.dif = scene.sphere[sid].dif;
		fragMat.spec = scene.sphere[sid].spec;
		fragMat.specExp = scene.sphere[sid].specExp;
		break;
	case RGRAY_TRIANGLE:
		obj = &scene.tri[ tid];
		uid = scene.tri[ tid].uid;
		fragPos = tip;
		fragNorm = scene.tri[tid].n;
		if( rd.dot( fragNorm) > 0) fragNorm = -1.0f * fragNorm; 
		fragMat.amb = scene.tri[ tid].amb;
		fragMat.dif = scene.tri[ tid].dif;
		fragMat.spec = scene.tri[ tid].spec;
		fragMat.specExp = scene.tri[ tid].specExp;
		break;
	default:
		return false;
	}

	/* If there is an object to trace */
	if( obj) {

		/*******************************************************
		 * Lighting (or shadowing) the fragment.
		 *******************************************************/
		rgRayLightFragment(
			fragPos,
			fragMat,
			fragNorm,
			colour,
			scene,
			uid);		/* UID to skip */


		/******************************************************
		 * Perform reflection if the fragment is reflective.
		 ******************************************************/
		if( obj->reflec > 0.0f) {
			rgVector rcol;
/*
			cerr << "fragNorm = " << fragNorm << "\n";
			cerr << "fragPos = " << fragPos << "\n";
			cerr << "rd = " << rd << "\n";
			cerr << "uid = " << uid << "\n";
			cerr << "depth = " << depth << "\n";
			cerr << "ref = " << rgReflectAbout(-1.0f*rd,fragNorm)
				<< "\n";
			exit(0);
*/
			if( rgRay_trace( fragPos,
				rgReflectAbout(-1.0f*rd, fragNorm), scene, rcol,
				uid, depth - 1)) {
//cerr << "hit!!!\n";
				//colour = (colour * (1.0f - obj->reflec)) +
				//	(rcol * obj->reflec);

				colour = colour +
					(rcol * obj->reflec);
			}
		}



		/******************************************************
		 * Perform refraction if the fragment is refractive.
		 ******************************************************/
		/*
		if( obj->transparent > 0.0f) {
			rgVector rcol;
			if( rgRay_trace( fragPos,
				
		}
		*/

		hit = true;
	}

	return hit;
}

rgFloat getSpecularScalar( rgVector liPos, rgVector fragPos, rgVector fragNorm,
	rgVector eyePos, rgFloat specExp)
{
	if( rgDotAngle( liPos - fragPos, fragNorm) >= RG_90_DEG_IN_RAD)
		return 0.0f;

	rgVector r = rgReflectAbout( (liPos - fragPos) + fragNorm, fragNorm);

	r = r + fragNorm;

	rgFloat ang = rgDotAngle( eyePos - fragPos, r);
	if( ang >= RG_90_DEG_IN_RAD) return 0.0f;

	return rgFloat( pow( cos( ang), specExp));
}

rgFloat getDiffuseScalar( rgVector liPos, rgVector fragPos, rgVector fragNorm)
{
	const rgFloat ang = rgDotAngle( liPos - fragPos, fragNorm);

	if( ang >= RG_90_DEG_IN_RAD) return 0.0f;
	
	return 1.0f - (ang/RG_90_DEG_IN_RAD);
}

bool rgRayInShadow( rgRayScene &scene, rgVector rp, rgVector rd,
	rgLONG p_uid_to_skip)
{
	rgVector c1;
	rgVector c2;

	int i;
	for( i = 0; i < scene.sphereCount; i++) {

		/* Skip sphere with the given uid (this prevents self
		 * shadowing, since spheres cannot shadow themselves)
		 */
		if( p_uid_to_skip != scene.sphere[i].uid) {
			if( sphereRayIntersection(
				scene.sphere[i].center,
				scene.sphere[i].radius,
				rp,
				rd,
				c1,
				c2))
			{
				return true;
			}
		}
	}

	for( i = 0; i < scene.triCount; i++) {

		if( p_uid_to_skip != scene.tri[i].uid) {
			if( rgRayTriangleIntersect(
				scene.tri[i].a,
				scene.tri[i].b,
				scene.tri[i].c,
				scene.tri[i].n,
				rp,
				rd,
				c1))
			{
				return true;
			}
		}
	}

	return false;
}

void rgRayFindClosestSphere( rgRayScene &scene, rgVector rp, rgVector rd,
	int &sid, rgVector &ip, rgLONG p_uid_to_skip)
{
	rgVector c1;
	rgVector c2;
	
	rgFloat dis = 32767.0f;
	rgFloat curDis;

	rgRaySphere *s = scene.sphere;

	sid = -1;
	int i;
	for( i = 0; i < scene.sphereCount; i++) {

		if( p_uid_to_skip != scene.sphere[i].uid) {

			if( sphereRayIntersection(
				s[i].center,
				s[i].radius,
				rp,
				rd,
				c1,
				c2))
			{
				const rgFloat k = (c1 - rp).length();
				curDis = (c2 - rp).length();

				if( k < curDis) {
				/* c1 is closer */

					if( k < dis) { /* If sphere is closer */
						dis = k;
						ip = c1;
						sid = i;
					}
				} else {
				/* c2 is closer. */

					if( curDis < dis)  {
					/* If sphere is closer */
						dis = curDis;
						ip = c2;
						sid = i;
					}
				}

			}

		}
	}
}

void rgRayFindClosestTriangle( rgRayScene &scene, rgVector rp, rgVector rd,
	int &tid, rgVector &ip, rgLONG p_uid_to_skip)
{
	rgVector cip;
	rgFloat dis = 32767.0f;

	rgRayTriangle *t = scene.tri;

	tid = -1;
	int i;
	for( i = 0; i < scene.triCount; i++) {

		if( p_uid_to_skip != scene.tri[i].uid) {

			if( rgRayTriangleIntersect(
				t[i].a, t[i].b, t[i].c, t[i].n, rp, rd, cip))
			{
				const rgFloat d = (cip - rp).length();
				if( d < dis) {
					dis = d;
					tid = i;
					ip = cip;
				}
			}

		}
	}
}


int rgRayLightFragment( rgVector fragPos, rgRayColourMaterial fragMat,
	rgVector fragNorm,
	rgVector &colour, rgRayScene &scene, rgLONG p_uid_to_skip)
{

	rgVector liCol;		/* Light Colour */
	rgVector sr;		/* Shadow Ray */

	int lightsUsed = 0;
	bool hit;

	/* Initialize final fragment colour (set to black). */
	colour = rgVector();

	int i;
	for( i = 0; i < scene.lightCount; i++) {

		/* Calculate the shadow ray for the current light */
		sr = (scene.light[i].position - fragPos);
		sr.normalize();

		/***********************************************************
		 * Check if current light is occluded by an object in the
		 * scene.
		 ***********************************************************/
		hit = rgRayInShadow(
			scene,
			fragPos,
			sr,
			p_uid_to_skip);

		/***********************************************************
		 * Do extra lighting calculations if the fragment is not in
		 * shadow. 
		 ***********************************************************/
		if( !hit) {
			liCol = scene.light[i].dif;
			liCol.x *= fragMat.dif.x;
			liCol.y *= fragMat.dif.y;
			liCol.z *= fragMat.dif.z;

			colour = colour + (liCol * getDiffuseScalar(
				scene.light[i].position,
				fragPos,
				fragNorm));

			liCol = scene.light[i].spec;
			liCol.x *= fragMat.spec.x;
			liCol.y *= fragMat.spec.y;
			liCol.z *= fragMat.spec.z;

			colour = colour + (liCol *
				getSpecularScalar(
				scene.light[i].position,
				fragPos,
				fragNorm,
				rgVector() ,
				fragMat.specExp
				));

			colour.x += fragMat.amb.x * scene.light[i].amb.x; 
			colour.y += fragMat.amb.y * scene.light[i].amb.y; 
			colour.z += fragMat.amb.z * scene.light[i].amb.z; 

			lightsUsed++;
		}
	}

	if( scene.lightCount > 1)
		colour = colour * (1.0f/rgFloat(scene.lightCount));


	return lightsUsed;
}














